"""Authorization related exceptions"""


class AuthorizationException(Exception):
    """Raised when an authorization is denied"""


class UndefinedAction(AuthorizationException):
    """Raised when attempting to authorize an undefined action"""



"""Basic building block of the Entitled library.

Rules are essentially wrappers around boolean functions that determines a relationship
between an actor and a resourc.

Resource can essentially be any entity in an application.
Actors represents any entity that, in an application, can act upon a resource.
"""

from typing import Any, Callable, ClassVar, Generic, Protocol, TypeVar

from entitled import exceptions

Resource = TypeVar("Resource", contravariant=True)


class RuleProtocol(Protocol[Resource]):
    """Defines valid functions for rules"""

    def __call__(
        self,
        actor: Any,
        resource: Resource | type[Resource],
        context: dict[str, Any] | None = None,
    ) -> bool: ...


class Rule(Generic[Resource]):
    """Base class for rules

    Args:
        name (str): The name given to this rule.
        rule_function (Callable): The underlying boolean function.
    """

    _registry: ClassVar[dict[str, "Rule[Any]"]] = {}

    @classmethod
    def clear_registry(cls) -> None:
        cls._registry = {}

    @classmethod
    def get_rule(cls, name: str) -> "Rule[Resource] | None":
        return Rule._registry[name] if name in Rule._registry else None

    def __init__(self, name: str, rule_function: RuleProtocol[Resource]):
        self.name = name
        self.rule = rule_function
        self.__register()

    def __register(self) -> None:
        if self.name in Rule._registry:
            print(f"A rule identified by '{self.name}' already exists.")
            return

        Rule._registry[self.name] = self

    def __call__(
        self,
        actor: Any,
        resource: Resource | type[Resource],
        context: dict[str, Any] | None = None,
    ) -> bool:
        if not context:
            context = {}
        return self.rule(actor, resource, context)

    def authorize(
        self,
        actor: Any,
        resource: Resource | type[Resource],
        context: dict[str, Any] | None = None,
    ) -> bool:
        if not self(actor, resource, context):
            raise exceptions.AuthorizationException("Unauthorized")
        return True

    def allows(
        self,
        actor: Any,
        resource: Resource | type[Resource],
        context: dict[str, Any] | None = None,
    ) -> bool:
        return self(actor, resource, context)


def rule(name: str) -> Callable[[RuleProtocol[Resource]], RuleProtocol[Resource]]:
    def wrapped(func: RuleProtocol[Resource]):
        _ = Rule[Resource](name, func)
        return func

    return wrapped



"""Centralization of all decision-making processes on a single decision point"""

import importlib.util
import pathlib
import types
from typing import Any

from entitled import policies


class Client:
    "The Client class for decision-making centralization."

    def __init__(self, base_path: str | None = None):
        self._policy_registrar: dict[type, policies.Policy[Any]] = {}
        self._load_path = None
        if base_path:
            self._load_path = pathlib.Path(base_path)
            self.load_policies_from_path(self._load_path)

    def authorize(
        self,
        action: str,
        actor: Any,
        resource: Any,
        context: dict[str, Any] | None = None,
    ) -> bool:
        policy = self._policy_lookup(resource)
        return policy.authorize(action, actor, resource, context)

    def allows(
        self,
        action: str,
        actor: Any,
        resource: Any,
        context: dict[str, Any] | None = None,
    ) -> bool:
        policy = self._policy_lookup(resource)
        return policy.allows(action, actor, resource, context)

    def grants(
        self,
        actor: Any,
        resource: Any,
        context: dict[str, Any] | None = None,
    ) -> dict[Any, bool]:
        policy = self._policy_lookup(resource)
        return policy.grants(actor, resource, context)

    def register(self, policy: policies.Policy[Any]):
        if hasattr(policy, "__orig_class__"):
            resource_type = getattr(policy, "__orig_class__").__args__[0]
            if resource_type not in self._policy_registrar:
                self._policy_registrar[resource_type] = policy
            else:
                raise ValueError(
                    "A policy is already registered for this resource type"
                )
        else:
            raise AttributeError(f"Policy {policy} is incorrectly defined")

    def reload_registrar(self):
        if self._load_path is not None:
            self.load_policies_from_path(self._load_path)

    def load_policies_from_path(self, path: pathlib.Path):
        for file_path in path.glob("*.py"):
            print(file_path)
            mod_name = file_path.stem
            full_module_name = ".".join(file_path.parts[:-1] + (mod_name,))
            spec = importlib.util.spec_from_file_location(full_module_name, file_path)
            if spec:
                module = importlib.util.module_from_spec(spec)
                if spec.loader:
                    try:
                        spec.loader.exec_module(module)
                    except Exception as e:
                        raise e

                    self._register_from_module(module)

    def _register_from_module(self, module: types.ModuleType):
        for attribute_name in dir(module):
            attr = getattr(module, attribute_name)
            if isinstance(attr, policies.Policy):
                try:
                    self.register(attr)
                except (ValueError, AttributeError):
                    pass

    def _policy_lookup(self, resource: Any) -> policies.Policy[Any]:
        lookup_key = resource if isinstance(resource, type) else type(resource)

        if lookup_key not in self._policy_registrar:
            raise ValueError("No policy registered for this resource type")

        return self._policy_registrar[lookup_key]



"""Grouping of authorization rules around a particular resource type"""

from typing import Any, Callable, Generic, TypeVar

from entitled import exceptions
from entitled.rules import Rule, RuleProtocol

T = TypeVar("T")


class Policy(Generic[T]):
    """A grouping of rules refering the given resource type."""

    def __init__(
        self,
        label: str | None = None,
        rules: dict[str, list[Rule[T]]] | None = None,
    ):
        self._registry: dict[
            str,
            list[Rule[T]],
        ] = {}
        self.label = label

        if not rules:
            rules = {}

        for action, rule in rules.items():
            self.__register(action, *rule)

    def rule(self, name: str) -> Callable[[RuleProtocol[T]], RuleProtocol[T]]:
        def wrapped(func: RuleProtocol[T]):
            rule_name = name
            if self.label is not None:
                rule_name = self.label + ":" + rule_name
            new_rule = Rule[T](rule_name, func)
            self.__register(name, new_rule)
            return func

        return wrapped

    def __register(self, action: str, *rules: Rule[T]):
        if action not in self._registry:
            self._registry[action] = [*rules]

    def grants(
        self, actor: Any, resource: T | type[T], context: dict[str, Any] | None = None
    ) -> dict[Any, bool]:
        return {
            action: self.allows(action, actor, resource, context)
            for action in self._registry
        }

    def allows(
        self,
        action: str,
        actor: Any,
        resource: T | type[T],
        context: dict[str, Any] | None = None,
    ) -> bool:
        try:
            return self.authorize(action, actor, resource, context)
        except exceptions.AuthorizationException:
            return False

    def authorize(
        self,
        action: str,
        actor: Any,
        resource: T | type[T],
        context: dict[str, Any] | None = None,
    ) -> bool:
        if action not in self._registry:
            raise exceptions.UndefinedAction(
                f"Action <{action}> undefined for this policy"
            )

        if not any(rule(actor, resource, context) for rule in self._registry[action]):
            raise exceptions.AuthorizationException("Unauthorized")

        return True



"""Importing some base classes at package root level for convenience"""

from entitled.client import Client
from entitled.policies import Policy
from entitled.rules import Rule

__all__ = ["Rule", "Policy", "Client"]
